from __future__ import annotations

import asyncio
import json
import ssl
from asyncio import CancelledError, TimeoutError, create_task, open_connection, wait_for
from contextlib import suppress
from dataclasses import dataclass, field
from logging import getLogger
from math import inf
from ssl import OP_NO_TLSv1, OP_NO_TLSv1_1, SSLError, create_default_context
from time import time
from typing import Dict, List, Optional
from urllib.parse import urlparse

import h2.config
import h2.connection
import h2.events
import h2.exceptions
import h2.settings

from .errors import Blocked, Closed, FormatError, ResponseTooLarge, StreamReset, Timeout

# Apple limits APN payload (data) to 4KB or 5KB, depending.
# Request header is not subject to flow control in HTTP/2
# Data is subject to framing and padding, but those are minor.
MAX_NOTIFICATION_PAYLOAD_SIZE = 5120
REQUIRED_FREE_SPACE = 6000
# OK response is empty
# Error response is short json, ~30 bytes in size
MAX_RESPONSE_SIZE = 2 ** 16
# Inbound connection flow control window
# It's quite arbitrary, guided by:
# * concurrent requests limit, server limit being 1000 today
# * expected response size, see above
CONNECTION_WINDOW_SIZE = 1000 * MAX_RESPONSE_SIZE
# Connection establishment safety time limits
CONNECTION_TIMEOUT = 5
TLS_TIMEOUT = 5
logger = getLogger(__package__)


@dataclass(eq=False)
class Connection:
    """Encapsulates a single HTTP/2 connection to the APN server.

    Use `Connection.create(...)` to make connections.

    Example use:

        conn = await Connection.create(...)
        try:
            await conn.post(request)
        finally:
            await conn.close()

    Connection states:
    * new (not connected)
    * starting
    * active
    * graceful shutdown (to do: https://github.com/python-hyper/hyper-h2/issues/1181)
    * closing
    * closed
    """

    host: str
    port: int
    protocol: h2.connection.H2Connection
    read_stream: asyncio.StreamReader
    write_stream: asyncio.StreamWriter
    should_write: asyncio.Event = field(init=False)
    channels: Dict[int, Channel] = field(default_factory=dict)
    reader: asyncio.Task = field(init=False)
    writer: asyncio.Task = field(init=False)
    closing: bool = False
    closed: bool = False
    outcome: Optional[str] = None
    max_concurrent_streams: int = 100  # initial per RFC7540#section-6.5.2
    last_stream_id_got: int = -1
    last_stream_id_sent: int = -1  # client streams are odd

    @classmethod
    async def create(
        cls, origin: str, ssl: Optional[ssl.SSLContext] = None
    ) -> Connection:
        """Connect to `origin` and return a Connection"""
        url = urlparse(origin)
        if (
            url.scheme != "https"
            or not url.hostname
            or url.username
            or url.password
            or url.path
            or url.params
            or url.query
            or url.fragment
        ):
            raise ValueError("Origin must be https://<host>[:<port>]")

        host = url.hostname
        port = url.port or 443

        ssl_context = ssl if ssl else create_ssl_context()
        if (
            OP_NO_TLSv1 not in ssl_context.options  # type: ignore # https://github.com/python/typeshed/issues/3920
            or OP_NO_TLSv1_1 not in ssl_context.options  # type: ignore
        ):
            raise ValueError("SSL Context cannot allow TLS 1.0 or 1.1")

        # https://bugs.python.org/issue40111 validate context h2 alpn

        protocol = h2.connection.H2Connection(
            h2.config.H2Configuration(client_side=True, header_encoding="utf-8")
        )
        protocol.local_settings = h2.settings.Settings(
            client=True,
            initial_values={
                # Apple server settings:
                # HEADER_TABLE_SIZE 4096
                # MAX_CONCURRENT_STREAMS 1000
                # INITIAL_WINDOW_SIZE 65535
                # MAX_FRAME_SIZE 16384
                # MAX_HEADER_LIST_SIZE 8000
                h2.settings.SettingCodes.ENABLE_PUSH: 0,
                h2.settings.SettingCodes.MAX_CONCURRENT_STREAMS: 2 ** 20,
                h2.settings.SettingCodes.MAX_HEADER_LIST_SIZE: 2 ** 16 - 1,
                h2.settings.SettingCodes.INITIAL_WINDOW_SIZE: MAX_RESPONSE_SIZE,
            },
        )

        protocol.initiate_connection()
        protocol.increment_flow_control_window(CONNECTION_WINDOW_SIZE)

        read_stream, write_stream = await wait_for(
            open_connection(
                host, port, ssl=ssl_context, ssl_handshake_timeout=TLS_TIMEOUT
            ),
            CONNECTION_TIMEOUT,
        )
        try:
            info = write_stream.get_extra_info("ssl_object")
            if not info:
                raise Closed("Failed TLS handshake")
            proto = info.selected_alpn_protocol()
            if proto != "h2":
                raise Closed("Failed to negotiate HTTP/2")
        except Closed:
            write_stream.close()
            with suppress(SSLError, ConnectionError):
                await write_stream.wait_closed()
            raise

        # FIXME we could wait for settings frame from the server,
        # to tell us how much we can actually send, as initial window is small
        return cls(host, port, protocol, read_stream, write_stream)

    def __post_init__(self):
        self.should_write = asyncio.Event()
        self.should_write.set()
        self.reader = create_task(self.background_read(), name="bg-read")
        self.writer = create_task(self.background_write(), name="bg-write")

    async def post(self, request: "Request") -> "Response":
        """Post the `request` on the connection"""
        if len(request.body) > MAX_NOTIFICATION_PAYLOAD_SIZE:
            raise ValueError("Request body is too large")

        request.get_time_left_or_fail()
        if self.closing or self.closed:
            raise Closed(self.outcome)
        if self.blocked:
            raise Blocked()

        try:
            stream_id = self.protocol.get_next_available_stream_id()
        except h2.exceptions.NoAvailableStreamIDError:
            self.closing = True
            if not self.outcome:
                self.outcome = "Exhausted"
            raise Closed(self.outcome)

        assert stream_id not in self.channels

        self.last_stream_id_got = stream_id

        self.channels[stream_id] = channel = Channel()
        self.protocol.send_headers(
            stream_id, request.header_with(self.host, self.port), end_stream=False
        )
        self.protocol.send_data(stream_id, request.body, end_stream=True)
        self.should_write.set()

        try:
            while not self.closed:
                remaining = request.get_time_left_or_fail()
                channel.wakeup.clear()
                with suppress(TimeoutError):
                    await wait_for(channel.wakeup.wait(), remaining)
                for event in channel.events:
                    if isinstance(event, h2.events.ResponseReceived):
                        channel.header = dict(event.headers)
                    elif isinstance(event, h2.events.DataReceived):
                        channel.body += event.data
                        if len(channel.body) >= MAX_RESPONSE_SIZE:
                            raise ResponseTooLarge(f"Larger than {MAX_RESPONSE_SIZE}")
                    elif isinstance(event, h2.events.StreamEnded):
                        return Response.new(channel.header, channel.body)
                    elif isinstance(event, h2.events.StreamReset):
                        raise StreamReset()
                del channel.events[:]
            raise Closed(self.outcome)
        finally:
            # FIXME reset the stream, if:
            # * connection is still alive, and
            # * this stream did not end yet
            # Must be very careful not to break the connection
            # self.protocol.reset_stream(stream_id, 0)
            # self.should_write.set()
            del self.channels[stream_id]

    async def close(self):
        """Terminate the connection and free up the resources"""
        self.closing = True
        if not self.outcome:
            self.outcome = "Closed"
        try:
            # FIXME distinguish between cancellation and context exception
            if self.writer:
                self.writer.cancel()
                with suppress(CancelledError):
                    await self.writer

            if self.reader:
                self.reader.cancel()
                with suppress(CancelledError):
                    await self.reader

            self.closed = True
            self.should_write.set()

            # at this point, we must release or cancel all pending requests
            for channel in self.channels.values():
                channel.wakeup.set()

            self.write_stream.close()
            with suppress(SSLError, ConnectionError):
                await self.write_stream.wait_closed()
        finally:
            self.closed = True
            self.should_write.set()

    @property
    def state(self):
        return (
            "closed"
            if self.closed
            else "closing"
            if self.closing
            else "active"
            if self.writer
            else "starting"
            if self.should_write
            else "new"
        )

    @property
    def inflight(self):
        """Count of the requests that were sent out and are awaiting server response."""
        return self.pending - self.buffered

    @property
    def buffered(self):
        """Count of the requests that we are still to send out."""
        return (self.last_stream_id_got - self.last_stream_id_sent) // 2

    @property
    def pending(self):
        """Total count of pending requests."""
        return len(self.channels)

    @property
    def blocked(self):
        """Is this connection unable to process more requests, either for now or permanently?"""
        return (
            self.closing
            or self.closed
            or self.protocol.outbound_flow_control_window <= REQUIRED_FREE_SPACE
            # FIXME accidentally quadratic: .openxx iterates over all streams
            # could be kinda fixed by caching with clever invalidation...
            or self.protocol.open_outbound_streams >= self.max_concurrent_streams
        )

    async def background_read(self):
        try:
            while not self.closed:
                data = await self.read_stream.read(2 ** 16)
                if not data:
                    raise ConnectionError("Server closed the connection")

                for event in self.protocol.receive_data(data):
                    logger.debug("APN: %s", event)
                    stream_id = getattr(event, "stream_id", 0)
                    error = getattr(event, "error_code", None)
                    channel = self.channels.get(stream_id)

                    if isinstance(event, h2.events.RemoteSettingsChanged):
                        m = event.changed_settings.get(
                            h2.settings.SettingCodes.MAX_CONCURRENT_STREAMS
                        )
                        if m:
                            self.max_concurrent_streams = m.new_value
                    elif isinstance(event, h2.events.ConnectionTerminated):
                        # When Apple is not happy with the whole connection,
                        # it sends smth like {"reason": "BadCertificateEnvironment"}
                        # Catch it here, so that connection pool can be invalidated.
                        self.closing = True
                        if not self.outcome:
                            if event.additional_data:
                                try:
                                    self.outcome = json.loads(
                                        event.additional_data.decode("utf-8")
                                    )["reason"]
                                except Exception:
                                    self.outcome = str(event.additional_data[:100])
                            else:
                                self.outcome = str(event.error_code)
                        logger.info("Closing with %s", self.outcome)
                    elif not stream_id and error is not None:
                        logger.warning("Caught off guard: %s", event)
                        raise ConnectionError(str(error))
                    else:
                        if isinstance(event, h2.events.DataReceived):
                            # Stream flow control is responsibility of the channel.
                            # Connection flow control is handled here.
                            self.protocol.acknowledge_received_data(
                                event.flow_controlled_length, stream_id
                            )
                        if channel:
                            channel.events.append(event)
                            channel.wakeup.set()

                # Somewhat inefficient: wake up background writer just in case
                # it could be that we've received something that h2 needs to acknowledge
                self.should_write.set()

                # FIXME notify pool users about possible change to `.blocked`
                # * h2.events.WindowUpdated
                # * max_concurrent_streams change
                # * [maybe] starting a stream
                # * a stream getting closed (but not half-closed)
                # * closing / closed change
        except ConnectionError as e:
            if not self.outcome:
                self.outcome = str(e)
        except Exception:
            logger.exception("background read task died")
        finally:
            self.closing = self.closed = True
            self.should_write.set()
            for channel in self.channels.values():
                channel.wakeup.set()

    async def background_write(self):
        try:
            while not self.closed:
                data = None

                while not data:
                    if self.closed:
                        return

                    if data := self.protocol.data_to_send():
                        self.write_stream.write(data)
                        last_stream_id = self.last_stream_id_got
                        await self.write_stream.drain()
                        self.last_stream_id_sent = last_stream_id
                    else:
                        await self.should_write.wait()
                        self.should_write.clear()

        except (SSLError, ConnectionError) as e:
            if not self.outcome:
                self.outcome = str(e)
        except Exception:
            logger.exception("background write task died")
        finally:
            self.closing = self.closed = True


@dataclass
class Channel:
    wakeup: asyncio.Event = field(default_factory=asyncio.Event)
    events: List[h2.events.Event] = field(default_factory=list)
    header: Optional[dict] = None
    body: bytes = b""


@dataclass
class Request:
    header: tuple
    body: bytes
    deadline: float
    deadline_source: str

    def header_with(self, host: str, port: int) -> tuple:
        """Request header including :authority pseudo header field for target server"""
        return ((":authority", f"{host}:{port}"),) + self.header

    def get_time_left_or_fail(self) -> float:
        """Raises Timeout() if the request has timed out, or return remaining time"""
        if (remaining := self.deadline - time()) > 0:
            return remaining
        raise Timeout("Request timed out: %s", self.deadline_source)

    @classmethod
    def new(
        cls,
        path: str,
        header: Dict[str, str],
        data: dict,
        timeout: Optional[float] = 10,
        deadline: Optional[float] = None,
        expiration: Optional[float] = None,
    ) -> Request:
        if not path.startswith("/"):
            raise ValueError("Absolute URL path is required")

        deadlines = {"not set": inf}
        if timeout is not None:
            deadlines["timeout"] = time() + timeout
        if deadline is not None:
            deadlines["deadline"] = deadline
        if expiration is not None:
            deadlines["expiration"] = expiration
        deadline = min(deadlines.values())
        deadline_source = [name for name, v in deadlines.items() if v == deadline][0]

        request_header = (
            (":method", "POST"),
            (":scheme", "https"),
            (":path", path),
            *header.items(),
        )

        text = json.dumps(data, ensure_ascii=False, separators=(",", ":"))
        return cls(request_header, text.encode("utf-8"), deadline, deadline_source)


@dataclass
class Response:
    code: int
    header: Dict[str, str]
    data: Optional[dict]

    @classmethod
    def new(cls, header: Optional[dict], body: bytes) -> Response:
        head = {**(header or {})}
        code = int(head.pop(":status", "0"))
        try:
            return cls(code, head, json.loads(body) if body else None)
        except json.JSONDecodeError:
            raise FormatError(f"Not JSON: {body[:20]!r}")

    @property
    def apns_id(self) -> Optional[str]:
        return self.header.get("apns-id", None)

    @property
    def reason(self) -> Optional[str]:
        """Response JSON 'reason' value, used in error responses."""
        return self.data.get("reason") if self.data else None


def create_ssl_context() -> ssl.SSLContext:
    """A basic SSL context suitable for HTTP/2 and APN."""
    context = create_default_context()
    context.options |= OP_NO_TLSv1
    context.options |= OP_NO_TLSv1_1
    context.set_alpn_protocols(["h2"])
    return context



from typing import Any, Dict, Optional, Type


class APNSError(Exception):
    pass


class Blocked(APNSError):
    """This connection can't send more data at this point, can try later."""


class Closed(APNSError):
    """This connection is now closed, try another."""


class Timeout(APNSError):
    """The request deadline has passed."""


class FormatError(APNSError):
    """Response was weird."""


class ResponseTooLarge(APNSError):
    """Server response was larger than allowed."""


class StreamReset(APNSError):
    pass


class ResponseError(APNSError):
    codename: str

    def __init__(self, reason: str, apns_id: Optional[str]):
        self.reason = reason
        self.apns_id = apns_id
        super().__init__(reason)


class UnknownResponseError(ResponseError):
    codename = "!unknown"


CODES: Dict[str, Type[ResponseError]] = {}


def create(codename: str) -> Type[ResponseError]:
    cls: Type[ResponseError] = type(codename, (ResponseError,), {"codename": codename})
    CODES[codename] = cls
    return cls


BadCollapseId = create("BadCollapseId")
BadDeviceToken = create("BadDeviceToken")
BadExpirationDate = create("BadExpirationDate")
BadMessageId = create("BadMessageId")
BadPriority = create("BadPriority")
BadTopic = create("BadTopic")
DeviceTokenNotForTopic = create("DeviceTokenNotForTopic")
DuplicateHeaders = create("DuplicateHeaders")
IdleTimeout = create("IdleTimeout")
MissingDeviceToken = create("MissingDeviceToken")
MissingTopic = create("MissingTopic")
PayloadEmpty = create("PayloadEmpty")
BadCertificate = create("BadCertificate")
BadCertificateEnvironment = create("BadCertificateEnvironment")
ExpiredProviderToken = create("ExpiredProviderToken")
Forbidden = create("Forbidden")
InvalidProviderToken = create("InvalidProviderToken")
MissingProviderToken = create("MissingProviderToken")
BadPath = create("BadPath")
MethodNotAllowed = create("MethodNotAllowed")
Unregistered = create("Unregistered")
PayloadTooLarge = create("PayloadTooLarge")
TooManyProviderTokenUpdates = create("TooManyProviderTokenUpdates")
TooManyRequests = create("TooManyRequests")
InternalServerError = create("InternalServerError")
ServiceUnavailable = create("ServiceUnavailable")
Shutdown = create("Shutdown")


def get(reason: Any, apns_id: Optional[str]) -> ResponseError:
    return CODES.get(reason, UnknownResponseError)(reason, apns_id)



from __future__ import annotations

import asyncio
import ssl
from asyncio import (
    CancelledError,
    Event,
    TimeoutError,
    create_task,
    gather,
    sleep,
    wait_for,
)
from contextlib import contextmanager, suppress
from dataclasses import dataclass, field
from itertools import count
from logging import getLogger
from random import shuffle
from time import time
from typing import Optional, Set

from .connection import Connection, Request, Response, create_ssl_context
from .errors import Blocked, Closed, FormatError, ResponseTooLarge, Timeout

logger = getLogger(__package__)


@dataclass(eq=False)
class Pool:
    """Simple fixed-size connection pool with automatic replacement

    Example use:

        pool = Pool.create(...)
        try:
            await pool.post(request)
        finally:
            await pool.close()

    """

    origin: str
    size: int
    ssl_context: ssl.SSLContext
    active: Set[Connection]
    dying: Set[Connection] = field(default_factory=set)
    closing: bool = False
    closed: bool = False
    errors: int = 0
    retrying: int = 0
    completed: int = 0
    outcome: Optional[str] = None
    maintenance: asyncio.Task = field(init=False)
    maintenance_needed: asyncio.Event = field(default_factory=asyncio.Event)

    @classmethod
    async def create(cls, origin: str, size=2, ssl=None) -> Pool:
        """Connect to `origin` and return a connection pool"""
        if size < 1:
            raise ValueError("Connection pool size must be strictly positive")
        ssl_context = ssl or create_ssl_context()
        connections = set(
            await gather(
                *(Connection.create(origin, ssl=ssl_context) for i in range(size))
            )
        )
        # FIXME run the hook / ensure no connection is dead
        return cls(origin, size, ssl_context, connections)

    def __post_init__(self):
        self.maintenance = create_task(self.maintain(), name="maintenance")

    async def post(self, request: "Request") -> "Response":
        """Post the `request` on a connection in this pool, with retries"""
        with self.count_requests():
            for delay in (10 ** i for i in count(-3, 0.5)):
                if self.closing:
                    raise Closed(self.outcome)

                try:
                    return await self.post_once(request)
                except Blocked:
                    pass

                if self.closing:
                    raise Closed(self.outcome)

                if request.get_time_left_or_fail() < delay:
                    raise Timeout("Request would time out awaiting retry")

                try:
                    self.retrying += 1
                    await sleep(delay)
                finally:
                    self.retrying -= 1

            assert False, "unreachable"

    async def close(self):
        """Terminate the connection pool and free up the resources"""
        self.closing = True
        if not self.outcome:
            self.outcome = "Closed"
        try:
            if self.maintenance:
                self.maintenance.cancel()
                with suppress(CancelledError):
                    await self.maintenance

            await gather(
                *(connection.close() for connection in self.active | self.dying)
            )
        finally:
            self.closed = True

    def __repr__(self):
        bits = [
            self.state,
            self.origin,
            f"active:{len(self.active)}",
            f"dying:{len(self.dying)}",
        ]
        if self.state != "closed":
            bits.append(f"buffered:{self.buffered}")
            bits.append(f"inflight:{self.inflight}")
        bits.append(f"retrying:{self.retrying}")
        bits.append(f"completed:{self.completed}")
        bits.append(f"errors:{self.errors}")
        return "<Pool %s>" % " ".join(bits)

    def resize(self, size: int):
        """Resize the connection pool

        Give the pool some time to grow/shrink to new `size` afterwards.
        """
        if size < 1:
            raise ValueError("Connection pool size must be strictly positive")
        self.size = size
        self.maintenance_needed.set()

    @property
    def state(self):
        return "closed" if self.closed else "closing" if self.closing else "active"

    @property
    def inflight(self):
        """Count of the requests that were sent out and are awaiting server response."""
        return sum(c.inflight for c in self.active | self.dying)

    @property
    def buffered(self):
        """Count of the requests that we are still to send out."""
        return sum(c.buffered for c in self.active | self.dying)

    @property
    def pending(self):
        """Total count of pending requests."""
        return sum(c.pending for c in self.active | self.dying) + self.retrying

    def termination_hook(self, connection: Connection):
        """
        A hook to terminate the pool if/when client certificate expires.

        If Apple is not happy with our client certificate, it will close individual
        connections with a JSON blob with a specific message. All connections in the
        pool share same ssl context, and thus same client certificate. If one
        connection is closed so, then the entire pool is done for.
        """
        if not self.outcome and connection.outcome == "BadCertificateEnvironment":
            self.closing = True
            self.outcome = connection.outcome

    async def maintain(self):
        while True:
            if self.closing or self.closed:
                return

            for connection in list(self.active):
                if connection.closing:
                    self.active.remove(connection)
                    self.dying.add(connection)
                    self.termination_hook(connection)

            while len(self.active) > self.size:
                connection = self.active.pop()
                connection.closing = True
                self.dying.add(connection)
                self.termination_hook(connection)

            for connection in list(self.dying):
                if connection.closed:
                    self.dying.remove(connection)
                    self.termination_hook(connection)
                elif not connection.channels:
                    self.dying.remove(connection)
                    try:
                        await connection.close()
                    finally:
                        self.termination_hook(connection)
                if self.closing or self.closed:
                    return

            while len(self.active) < self.size:
                if not await self.add_one_connection():
                    break
                if self.closing or self.closed:
                    return

            # FIXME wait for a trigger:
            # * some connection state has changed
            with suppress(TimeoutError):
                await wait_for(self.maintenance_needed.wait(), timeout=1)
            self.maintenance_needed.clear()

    async def add_one_connection(self):
        try:
            connection = await Connection.create(self.origin, ssl=self.ssl_context)
            self.active.add(connection)
            self.termination_hook(connection)
            return True
        except OSError as e:
            logger.error("%s", e)
        except Exception:
            logger.exception("Failed creating APN connection")

    @contextmanager
    def count_requests(self):
        try:
            yield
        except:
            self.errors += 1
            raise
        else:
            self.completed += 1

    async def post_once(self, request: "Request") -> "Response":
        # FIXME ideally, follow weighted round-robin discipline:
        # * generally allocate requests evenly across connections
        # * but keep load for few last connections lighter
        #   to prevent all connections expiring at once
        # * ideally track connection backlog
        active = list(self.active)
        shuffle(active)
        for connection in active:
            if self.closing:
                raise Closed(self.outcome)
            if connection.closed:
                continue
            try:
                return await connection.post(request)
            except (Blocked, Closed):
                pass
        else:
            raise Blocked()



from enum import Enum

import attr


@attr.s
class Server:
    host = attr.ib()
    port = attr.ib()


Production = Server("api.push.apple.com", 443)
ProductionAltPort = attr.evolve(Production, port=2197)
Development = Server("api.development.push.apple.com", 443)
DevelopmentAltPort = attr.evolve(Development, port=2197)


class Priority(Enum):
    immediately = 10
    normal = 5



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aapns']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.3.0,<20.0.0', 'h2>=3.2.0,<4.0.0']

extras_require = \
{'cli': ['click>=7.0,<8.0']}

entry_points = \
{'console_scripts': ['aapns = aapns.cli:main']}

setup_kwargs = {
    'name': 'aapns',
    'version': '20.4b1',
    'description': 'Asynchronous Apple Push Notification Service Client',
    'long_description': '# AAPNS\n\n[![CircleCI](https://circleci.com/gh/HENNGE/aapns.svg?style=svg)](https://circleci.com/gh/HENNGE/aapns)\n[![Documentation Status](https://readthedocs.org/projects/aapns/badge/?version=latest)](http://aapns.readthedocs.io/en/latest/?badge=latest)\n\nAsynchronous Apple Push Notification Service client.\n\n* Requires TLS 1.2 or better\n* Requires Python 3.8 or better\n\n## Quickstart\n\n```python\nfrom aapns.api import create_client\nfrom aapns.config import Priority, Production\nfrom aapns.models import Notification, Alert, Localized\n\nasync def send_hello_world():\n    client = await create_client(\'/path/to/push/cert.pem\', Production)\n    apns_id = await client.send_notification(\n        \'my-device-token\',\n        Notification(\n            alert=Alert(\n                body=Localized(\n                    key=\'Hello World!\',\n                    args=[\'foo\', \'bar\']\n                ),\n            ),\n            badge=42\n        ),\n        priority=Priority.immediately\n    )\n    print(f\'Sent push notification with ID {apns_id}\')\n    await client.close()\n```\n\n### On idempotency\n\nOr how to ensure that a notification is displayed only once in the presence of network errors?\n\nIf the underlying TCP connection is broken or times out, there may be some notifications still in flight. In that case it\'s impossible to tell whether the notification was not yet delivered to the Apple server, or it was delivered but Apple server response was not yet delivered to back to your client.\n\nAdditionally, the server may try to shut down an HTTP/2 connection gracefully, for example for server maintenance or upgrade. In that case, due to https://github.com/python-hyper/hyper-h2/issues/1181, it is not possible to tell which of the in-flight requests will be completed by the server.\n\nThe Apple push notification protocol provides a header field `apns-collapse-id`. In the simple use-case, it\'s recommended to set this field, for example to a random value:\n\n```py\nawait client.send_notification(\n    ...,\n    collapse_id=str(random.random()),\n)\n```\n\nThen, should `aapns` need to retransmit the request, and retransmit "one time too many", the end user will only see a single notification.\n\nHowever, if you are rely on notification collapse mechanism, if `aapns` retransmits, the notifications may arrive out of order, and the end-user may ultimately see stale data.\n\n### Mid level API\n\nUse this API to maintain a fixed size connection pool and gain automatic retries with exponential back-off. A connection pool can handle up to `1000 * size` (current Apple server limit) concurrent requests and practically unlimited dumb queue of requests should concurrency limit be exceeded. It is thus suitable for bursty traffic.\n\nUse this API to send notification en masse or generic RPC-like communication over HTTP/2.\n\n```py\nfrom aapns.errors import APNSError, Closed, Timeout\nfrom aapns.pool import create_ssl_context, Pool, Request\n\n\nssl_context = create_ssl_context()\nssl_context.load_cert_chain(certfile=..., keyfile=...)\n\nreq = Request.new(\n    "/3/device/42...42",\n    {"apns-push-type": "alert", "apns-topic": "com.app.your"},\n    {"apns": {"alert": {"body": "Wakey-wakey, ham and bakey!"}}},\n    timeout=10,  # or the pool may retry forever\n)\n\nasync with Pool(\n        "https://api.development.push.apple.com",\n        size=10,  # default\n        ssl=ssl_context) as pool:\n    try:\n        resp = await pool.post(req)\n        assert resp.code == 200\n    except Timeout:\n        ...  # the notification has expired\n    except Closed:\n        ...  # the connection pool is done, e.g. if client certificate has expired\n    except APNSError:\n        ...  # rare\n```\n\n### Low level API\n\nUse this API if you want close control of a single connection to the server. A connection can handle up to `1000` concurrent requests (current Apple server limit) and up to `2**31` requests in total (HTTP/2 protocol limit).\n\nThis would be a good start for token authentication, https://github.com/HENNGE/aapns/issues/19.\n\n```py\nfrom aapns.errors import APNSError, Blocked, Closed, Timeout\nfrom aapns.connection import create_ssl_context, Connection, Request\n\n\nssl_context = create_ssl_context()\nssl_context.load_cert_chain(certfile=..., keyfile=...)\n\nreq = Request.new(\n    "/3/device/42...42",\n    {"apns-push-type": "alert", "apns-topic": "com.app.your"},\n    {"apns": {"alert": {"body": "Wakey-wakey, ham and bakey!"}}},\n    timeout=10)\n\nasync with Connection(\n        "https://api.development.push.apple.com",\n        ssl=ssl_context) as conn:\n    try:\n        resp = await conn.post(req)\n        assert resp.code == 200\n    except Blocked:\n        ...  # the connection is busy, try again later\n    except Closed:\n        ...  # the connection is no longer usable\n    except Timeout:\n        ...  # the notification has expired\n    except APNSError:\n        ...  # rare\n```\n\n### Technical notes\n\nRationale for using raw https://github.com/python-hyper/hyper-h2 rather than an existing library, like https://github.com/encode/httpx.\n\nContrast push notification use-case vs. generic, browser-like use-case:\n\n| feature                   | push-like | browser-like                        |\n|:--------------------------|:----------|:------------------------------------|\n| request size              | tiny      | small or large                      |\n| request method            | `POST`    | `OPTIONS,HEAD,GET,PUT,POST`,custom  |\n| response size             | tiny      | small, large, giant, streamed       |\n| server push               | no        | possible                            |\n| concurrent per connection | `1000`    | dozens                              |\n| total per connection      | millions  | dozens                              |\n| retryable                 | all       | idempotent verbs, graceful shutdown |\n| servers                   | `1`       | many                                |\n| authorisation             | client cert or token | none, token, other       |\n\n* Apple server sets max concurrent requests to `1000` and push requests are small (5KB max), thus TCP send buffer will be quite small, thus:\n  * we\'re not setting `TCP_NOTSENT_LOWAT`\n  * we\'re not checking `SO_NWRITE/SIOCOUTQ`\n* APN server is available on IPv4 only today, thus we don\'t worry about happy eyeballs\n',
    'author': 'Jonas Obrist',
    'author_email': 'jonas.obrist@hennge.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/HENNGE/aapns',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
